"""Calculate expected CLs values with hypothesis tests."""
from __future__ import annotations

__all__ = ["make_hypotest"]

from ..fit import constrained_fit
from ..fit.minuit_transforms import to_bounded_vec, to_inf_vec
from .._types import ArrayDevice

from typing import Any, Callable
import jax
import jax.numpy as jnp
import pyhf

pyhf.set_backend("jax")
jax.config.update("jax_enable_x64", True)


def make_hypotest(
    model_maker: Callable[[Any], tuple[Any, ArrayDevice]],
    solver_kwargs: dict[str, Any] = dict(pdf_transform=True),
) -> Callable[[ArrayDevice, float, list[str]], dict[str, float]]:
    """Instatiate a hypotheses test based on a model maker.

    Parameters
    ----------
    model_maker: function that takes model params and returns the tuple
    [model, background-only parameters] when called.
    solver_kwargs: keyword arguments passed to constrained_fit.

    Returns
    -------
    hypotest: function that performs a hypothesis test using the profile
    likelihood as a test statistic. calculates CLs values in the expected
    case of observing the nominal signal + background yields.
    """

    @jax.jit
    def hypotest(
        hyperpars: ArrayDevice, test_mu: float, metrics: list[str] = ["CLs"]
    ) -> dict[str, float]:
        # g_fitter = global_fit(model_maker, **solver_kwargs)
        c_fitter = constrained_fit(model_maker, **solver_kwargs)

        m, bonlypars = model_maker(hyperpars)
        exp_data = m.expected_data(bonlypars)
        bounds = jnp.array(m.config.suggested_bounds())
        # map these
        initval = jnp.asarray([test_mu, 1.0])
        transforms = solver_kwargs.get("pdf_transform", False)
        if transforms:
            initval = to_inf_vec(initval, bounds)

        # the constrained fit
        numerator = (
            to_bounded_vec(c_fitter(initval, (hyperpars, test_mu)), bounds)
            if transforms
            else c_fitter(initval, (hyperpars, test_mu))
        )

        # don't have to fit these -- we know them for expected limits!
        denominator = bonlypars

        # compute test statistic (lambda(µ))
        profile_likelihood = -2 * (
            m.logpdf(numerator, exp_data)[0] - m.logpdf(denominator, exp_data)[0]
        )

        # in exclusion fit zero out test stat if best fit µ^ is larger than test µ
        muhat = denominator[0]
        sqrtqmu = jnp.sqrt(jnp.where(muhat < test_mu, profile_likelihood, 0.0))
        CLsb = 1 - pyhf.tensorlib.normal_cdf(sqrtqmu)
        altval = 0
        CLb = 1 - pyhf.tensorlib.normal_cdf(altval)
        CLs = CLsb / CLb

        pdict = dict(CLs=CLs, p_sb=CLsb, p_b=CLb, profile_likelihood=profile_likelihood)
        return {k: pdict[k] for k in metrics}

    return hypotest



"""in part taken from deepmind/chex while waiting for a better way to type jax."""

__all__ = ["Array", "ArrayNumpy", "ArrayDevice", "PRNGKey"]
import jax
import jax.numpy as jnp
import numpy as np

Array = jnp.ndarray
ArrayNumpy = np.ndarray
# Use this type for type annotation. For instance checking,  use
# `isinstance(x, jax.DeviceArray)`.
# `jax.interpreters.xla._DeviceArray` appears in jax > 0.2.5
if hasattr(jax.interpreters.xla, "_DeviceArray"):
    ArrayDevice = jax.interpreters.xla._DeviceArray
else:
    ArrayDevice = jax.interpreters.xla.DeviceArray
PRNGKey = Array



"""differentiable implementations of histograms."""

__all__ = ["hist_kde"]

from typing import Optional
import jax
import jax.numpy as jnp
import jax.scipy as jsc

from .._types import ArrayDevice


@jax.jit
def hist_kde(
    events: ArrayDevice,
    bins: ArrayDevice,
    bandwidth: Optional[float] = None,
    density: bool = False,
) -> ArrayDevice:
    """
    Differentiable implementation of a histogram using kernel density estimation.

    Parameters
    ----------
    events: (jax array-like) 1D data!
    bins: (jax array-like) intervals to calculate counts.
    bandwidth: (float) value that specifies the width of the individual
        distributions (kernels), whose cdfs are averaged over each bin. Defaults
        to Scott's rule -- the same as scipy's.
    density: (bool) whether or not to normalize the histogram to unit area.

    Returns
    -------
    counts: 1D array of binned counts
    """
    bandwidth = bandwidth or events.shape[-1] ** -0.25  # Scott's rule

    edge_hi = bins[1:]  # ending bin edges ||<-
    edge_lo = bins[:-1]  # starting bin edges ->||

    # get cumulative counts (area under kde) for each set of bin edges
    cdf_up = jsc.stats.norm.cdf(edge_hi.reshape(-1, 1), loc=events, scale=bandwidth)
    cdf_dn = jsc.stats.norm.cdf(edge_lo.reshape(-1, 1), loc=events, scale=bandwidth)
    # sum kde contributions in each bin
    counts = (cdf_up - cdf_dn).sum(axis=1)

    if density:  # normalize by bin width and counts for total area = 1
        db = jnp.array(jnp.diff(bins), float)  # bin spacing
        return counts / db / counts.sum(axis=0)

    return counts



"""instantiate routines for fitting a model given a fixed param of intrest."""
from __future__ import annotations

from typing import Any
from typing import Callable

import jax
import jax.experimental.optimizers as optimizers
import jax.numpy as jnp
from fax.implicit import two_phase_solver

from .._types import ArrayDevice
from .minuit_transforms import to_bounded_vec
from .minuit_transforms import to_inf


def constrained_fit(
    model_maker: Callable[[Any], tuple[Any, ArrayDevice]],
    pdf_transform: bool = False,
    default_rtol: float = 1e-10,
    default_atol: float = 1e-10,
    default_max_iter: int = int(1e7),
    learning_rate: float = 1e-6,
) -> Callable[[ArrayDevice, tuple[ArrayDevice, float]], ArrayDevice]:
    """Initialize a constrained fit of a model via gradient descent w/ adam optimizer.

    Parameters
    ----------
    model_maker: function that takes model params and returns the tuple
    [model, background-only parameters] when called.
    pdf_transform: whether to map the likelihood domain to [-inf, inf]
    during minimization.
    default_rtol, default_atol, default_max_iter: args to pass to fax's
    two-phase solver.
    learning_rate: controls step-size in grad descent. arg supplied to adam optimizer.

    Returns
    -------
    constrained_fitter: function that peforms the fit in a differentiable way
    via Christianson's two-phase method.
    Takes in (initial params, model hyperparams) as args.
    """
    adam_init, adam_update, adam_get_params = optimizers.adam(learning_rate)

    def make_model(
        hyper_pars: tuple[ArrayDevice, float]
    ) -> tuple[float, Callable[[ArrayDevice], float]]:

        model_pars, constrained_mu = hyper_pars
        m, bonlypars = model_maker(model_pars)

        bounds = jnp.array(m.config.suggested_bounds())
        constrained_mu = (
            to_inf(constrained_mu, bounds[0]) if pdf_transform else constrained_mu
        )

        exp_bonly_data = m.expected_data(bonlypars, include_auxdata=True)

        def expected_logpdf(
            pars: ArrayDevice,
        ) -> tuple[float]:  # maps pars to bounded space if pdf_transform = True

            return (
                m.logpdf(to_bounded_vec(pars, bounds), exp_bonly_data)
                if pdf_transform
                else m.logpdf(pars, exp_bonly_data)
            )

        def constrained_fit_objective(nuis_par: ArrayDevice) -> float:  # NLL
            pars = jnp.concatenate([jnp.asarray([constrained_mu]), nuis_par])
            return -expected_logpdf(pars)[0]

        return constrained_mu, constrained_fit_objective

    def constrained_bestfit_minimized(
        hyper_pars: tuple[ArrayDevice, float]
    ) -> Callable[[int, ArrayDevice], ArrayDevice]:
        mu, cnll = make_model(hyper_pars)

        def bestfit_via_grad_descent(
            i: int, param: ArrayDevice
        ) -> ArrayDevice:  # gradient descent
            _, np = param[0], param[1:]
            grads = jax.grad(cnll)(np)
            np = adam_get_params(adam_update(i, grads, adam_init(np)))
            param = jnp.concatenate([jnp.asarray([mu]), np])
            return param

        return bestfit_via_grad_descent

    constrained_solver = two_phase_solver(
        param_func=constrained_bestfit_minimized,
        default_rtol=default_rtol,
        default_atol=default_atol,
        default_max_iter=default_max_iter,
    )

    def constrained_fitter(
        init: ArrayDevice, hyper_pars: tuple[ArrayDevice, float]
    ) -> ArrayDevice:
        solve = constrained_solver(init, hyper_pars)
        return solve.value

    return constrained_fitter



"""implementations of scaling transforms to aid minimisation, taken from minuit."""

__all__ = ["to_bounded_vec", "to_bounded", "to_inf_vec", "to_inf"]

import jax.numpy as jnp
from .._types import ArrayDevice


def to_bounded_vec(param: ArrayDevice, bounds: ArrayDevice) -> ArrayDevice:
    """[-inf, inf] -> [bounds[:, 0], bounds[:, 1]]."""
    a, b = bounds[:, 0], bounds[:, 1]
    return a + (b - a) * 0.5 * (jnp.sin(param) + 1.0)


# [-inf, inf] -> [a,b]
def to_bounded(param: ArrayDevice, bounds: ArrayDevice) -> ArrayDevice:
    """[-inf, inf] -> [bounds[0], bounds[1]]."""
    a, b = bounds
    return a + (b - a) * 0.5 * (jnp.sin(param) + 1.0)


# [-inf, inf] <- [a,b] (vectors)
def to_inf_vec(param: ArrayDevice, bounds: ArrayDevice) -> ArrayDevice:
    """[bounds[:, 0], bounds[:, 1]] -> [-inf, inf]."""
    a, b = bounds[:, 0], bounds[:, 1]
    x = 2.0 * (param - a) / (b - a) - 1.0
    return jnp.arcsin(x)


# [-inf, inf] <- [a,b]
def to_inf(param: ArrayDevice, bounds: ArrayDevice) -> ArrayDevice:
    """[bounds[0], bounds[1]] -> [-inf, inf]."""
    a, b = bounds
    x = 2.0 * (param - a) / (b - a) - 1.0
    return jnp.arcsin(x)
